@@ -102,7 +102,7 @@ class GTAB:
             default_anchorbank = "google_anchorbank_geo=_timeframe=2019-01-01 2020-08-01.tsv"
             self.set_active_gtab(default_anchorbank)
 
-    # --- UTILITY METHODS --- 
+    # --- UTILITY METHODS ---
 
     def _print_and_log(self, text, verbose=True):
         if verbose:
@@ -110,7 +110,11 @@ class GTAB:
         self._log_con.write(text + '\n')
 
     def _make_file_suffix(self):
-        return "_".join([f"{k}={v}" for k, v in self.CONFIG['PYTRENDS'].items()])
+
+        return "_".join([f"{k}={v}" for k, v in self.CONFIG['PYTRENDS'].items()
+                         if not (k == "cat" and v == "0")]  # this if ensures compatibility with previous name formats,
+                        # after we allow people to search for categories
+                        )
 
     def _query_google(self, keywords=["Keywords"]):
         time.sleep(self.CONFIG['GTAB']['sleep'])
@@ -120,14 +124,20 @@ class GTAB:
         if len(keywords) > 5:
             raise ValueError("Number of keywords must be at most than 5.")
 
-        self.pytrends.build_payload(kw_list=keywords, **self.CONFIG['PYTRENDS'])
-        ret = self.pytrends.interest_over_time()
+        # avoids duplicate query
+        if len(keywords) == 2 and keywords[0] == keywords[1]:
+            self.pytrends.build_payload(kw_list=[keywords[0]], **self.CONFIG['PYTRENDS'])
+            ret = self.pytrends.interest_over_time()
+            ret.insert(loc=1, column="tmp", value=ret.iloc[:, 0])
+        else:
+            self.pytrends.build_payload(kw_list=keywords, **self.CONFIG['PYTRENDS'])
+            ret = self.pytrends.interest_over_time()
         return ret
 
     def _is_not_blacklisted(self, keyword):
         return keyword not in self.CONFIG['BLACKLIST']
 
-    def _check_keyword(self, keyword):
+    def _check_keyword(self, keyword, retry_on_429=False):
         time.sleep(self.CONFIG['GTAB']['sleep'])
         try:
             rez = self._query_google(keywords=keyword)
@@ -135,12 +145,17 @@ class GTAB:
         except ValueError as e:
             raise e
         except Exception as e:
-            self._print_and_log(f"\nBad keyword '{keyword}', because {str(e)}")
-            if "response" in dir(e):
+            if "response" in dir(e) and e.response is not None:
                 if e.response.status_code == 429:
-                    raise ConnectionError("Code 429: Query limit reached on this IP!")
-
-            return False
+                    if retry_on_429:
+                        input("Quota reached! Please change IP and press enter to continue.")
+                        return self._check_keyword(keyword)
+                    else:
+                        raise ConnectionError("Code 429: Query limit reached on this IP!")
+                self._print_and_log(f"\nBad keyword '{keyword}', because {e}")
+            else:
+                raise e
+        return False
 
     def _check_ts(self, ts):
         return ts.max() >= self.CONFIG['GTAB']['thresh_offline']
@@ -202,7 +217,7 @@ class GTAB:
 
         # old heuristic
         bad_kws2 = self._check_groups(gres)
-        bad_idxs2 = [bad_dict[kw][0] for kw in bad_kws2 if bad_dict[kw][0] < len(gres) - 4]
+        bad_idxs2 = [bad_dict[kw][0] for kw in bad_kws2 if kw in bad_dict and bad_dict[kw][0] < len(gres) - 4]
 
         ret = sorted(bad_idxs1 + bad_idxs2)
 
@@ -216,6 +231,8 @@ class GTAB:
 
         fpath = os.path.join(self.dir_path, "data", "internal", "google_results",
                              f"google_results_{self._make_file_suffix()}.pkl")
+        fpath_intermediate = os.path.join(self.dir_path, "data", "internal", "google_results",
+                                          f"intermediate_google_results_{self._make_file_suffix()}.pkl")
         fpath_keywords = os.path.join(self.dir_path, "data", "internal", "google_keywords",
                                       f"google_keywords_{self._make_file_suffix()}.pkl")
 
@@ -276,10 +293,18 @@ class GTAB:
             N = self.CONFIG['GTAB']['num_anchor_candidates']
             K = self.CONFIG['GTAB']['num_anchors']
 
+            fpath_intermediate_keywords = os.path.join(self.dir_path, "data", "internal", "google_keywords",
+                                                       f"intermediate_google_keywords_{self._make_file_suffix()}.pkl")
+
             # --- Get stratified samples, 1 per stratum. ---
             if not os.path.exists(fpath_keywords):
+                if os.path.exists(fpath_intermediate_keywords):
+                    self._print_and_log(f"Loading previously checked keywords from {fpath_intermediate_keywords}.")
+                    keywords_with_status = self._load_pickle_with_fallback(fpath_intermediate_keywords,
+                                                                           on_error_return={})
+                else:
+                    keywords_with_status = {}
 
-                self._print_and_log("Sampling keywords...")
                 random.seed(self.CONFIG['GTAB']['seed'])
                 np.random.seed(self.CONFIG['GTAB']['seed'])
 
@@ -290,8 +315,19 @@ class GTAB:
                     s1 = random.randint(start_idx, end_idx)
                     samples.append(self.ANCHOR_CANDIDATES[s1])
 
-                keywords = self.HITRAFFIC + samples
-                keywords = [k for k in tqdm(keywords, total=len(keywords)) if self._check_keyword(k)]
+                keyword_candidates = self.HITRAFFIC + samples
+                remaining_keywords = [kw for kw in keyword_candidates if kw not in keywords_with_status]
+                self._print_and_log(f"{len(keyword_candidates) - len(remaining_keywords)}/{len(keyword_candidates)} "
+                                    f"keyword candidates found in cache, {len(remaining_keywords)} still to check.")
+                self._print_and_log("Sampling keywords...")
+                for kw in tqdm(remaining_keywords, total=len(remaining_keywords)):
+                    try:
+                        keywords_with_status[kw] = self._check_keyword(kw, retry_on_429=True)
+                    except Exception as e:
+                        with open(fpath_intermediate_keywords, 'wb') as f_in:
+                            pickle.dump(keywords_with_status, f_in, protocol=4)
+                        raise e
+                keywords = [kw for kw in keyword_candidates if keywords_with_status[kw]]
 
             else:
                 print(f"Getting keywords from {fpath_keywords}")
@@ -302,26 +338,43 @@ class GTAB:
             with open(fpath_keywords, 'wb') as f_out_keywords:
                 pickle.dump(keywords, f_out_keywords, protocol=4)
 
+            if os.path.exists(fpath_intermediate_keywords):
+                os.remove(fpath_intermediate_keywords)
+
             keywords = np.array(keywords)
 
             # --- Query google on keywords and dump the file. ---
             self._print_and_log("Querying google...")
 
-            t_ret = dict()
+            if os.path.exists(fpath_intermediate):
+                self._print_and_log(f"Loading intermediate query results from {fpath_intermediate}.")
+                query_cache = self._load_pickle_with_fallback(fpath_intermediate, on_error_return={})
+            else:
+                query_cache = {}
+            t_ret = {}
             for i in tqdm(range(len(keywords) - 4)):
-
-                try:
-                    df_query = self._query_google(keywords=keywords[i:i + 5]).iloc[:, 0:5]
-                except ValueError as e:
-                    raise e
-                except Exception as e:
-                    if "response" in dir(e):
-                        if e.response.status_code == 429:
-                            input("Quota reached! Please change IP and press any key to continue.")
-                            df_query = self._query_google(keywords=keywords[i:i + 5]).iloc[:, 0:5]
+                kw_group = keywords[i:i + 5]
+                cache_key = tuple(kw_group)
+                if cache_key in query_cache:
+                    t_ret[i] = query_cache[cache_key]
+                else:
+                    try:
+                        df_query = self._query_google(keywords=kw_group).iloc[:, 0:5]
+                        query_cache[cache_key] = df_query
+                    except ValueError as e:
+                        with open(fpath_intermediate, 'wb') as f_out:
+                            pickle.dump(query_cache, f_out, protocol=4)
+                        raise e
+                    except Exception as e:
+                        with open(fpath_intermediate, 'wb') as f_out:
+                            pickle.dump(query_cache, f_out, protocol=4)
+                        if "response" in dir(e) and e.response is not None and e.response.status_code == 429:
+                            input("Quota reached! Please change IP and press enter to continue.")
+                            df_query = self._query_google(keywords=kw_group).iloc[:, 0:5]
+                            query_cache[cache_key] = df_query
                         else:
-                            self._print_and_log(str(e))
-                t_ret[i] = df_query
+                            raise e
+                    t_ret[i] = df_query
 
             self._print_and_log("Removing bad queries...")
             no_passes = 0
@@ -365,20 +418,26 @@ class GTAB:
                     # ...re-query the ones that do
                     kw_groups = get_kws(req_rng, new_kws)
                     for kw_group in tqdm(kw_groups):
-
-                        try:
-                            df_query = self._query_google(keywords=kw_group).iloc[:, 0:5]
-                        except ValueError as e:
-                            raise e
-                        except Exception as e:
-                            if "response" in dir(e):
-                                if e.response.status_code == 429:
-                                    input("Quota reached! Please change IP and press any key to continue.")
+                        cache_key = tuple(kw_group)
+                        if cache_key in query_cache:
+                            ret[idx_new] = query_cache[cache_key]
+                        else:
+                            try:
+                                df_query = self._query_google(keywords=kw_group).iloc[:, 0:5]
+                            except ValueError as e:
+                                with open(fpath_intermediate, 'wb') as f_out:
+                                    pickle.dump(query_cache, f_out, protocol=4)
+                                raise e
+                            except Exception as e:
+                                with open(fpath_intermediate, 'wb') as f_out:
+                                    pickle.dump(query_cache, f_out, protocol=4)
+                                if "response" in dir(e) and e.response is not None and e.response.status_code == 429:
+                                    input("Quota reached! Please change IP and press enter to continue.")
                                     df_query = self._query_google(keywords=kw_group).iloc[:, 0:5]
                                 else:
-                                    self._print_and_log(str(e))
-
-                        ret[idx_new] = df_query
+                                    raise e
+                            query_cache[cache_key] = df_query
+                            ret[idx_new] = df_query
                         idx_new += 1
 
                 if start_copy < len(t_ret):
@@ -400,7 +459,18 @@ class GTAB:
                 self._print_and_log(f"Saving google results as '{fpath}'...")
                 pickle.dump(ret, f_out, protocol=4)
 
-            return (ret)
+            if os.path.exists(fpath_intermediate):
+                os.remove(fpath_intermediate)
+
+            return ret
+
+    def _load_pickle_with_fallback(self, fpath, on_error_return=None):
+        try:
+            with open(fpath, 'rb') as f_in:
+                return pickle.load(f_in)
+        except Exception as e:
+            self._print_and_log(f"Load pickle from path {fpath} failed with cause: {e}, resorting to fallback.")
+            return on_error_return
 
     def _compute_hi_and_lo(self, max1, max2):
 
@@ -427,7 +497,7 @@ class GTAB:
 
         if lo2 < 0:
             lo2 = 0.0
-            
+
         return lo1, hi1, lo2, hi2
 
     def _compute_max_ratios(self, google_results):
@@ -584,27 +654,34 @@ class GTAB:
 
     def set_options(self, pytrends_config=None, gtab_config=None, conn_config=None, overwite_file: bool = False):
         """
-            Overwrites specified options. This can also be done manually by editing 'config_py.json' in the active directory.
+        Overwrites specified options. This can also be done manually by editing 'config_py.json' in the active
+        directory.
 
-                pytrends_config - a dictionary containing values to overrwite some of the configuration parameters for the pytrends library when 
-                building the payload. It consists of two parameters:
-                    - geo (str) - containing the two-letter ISO code of the country, e.g. "US", or "" empty string for global;
-                    - timeframe (str) - containing the timeframe in which to query, e.g. "2019-11-28 2020-07-28".
+            pytrends_config - a dictionary containing values to overrwite some of the configuration parameters for the
+            pytrends library when
+            building the payload. It consists of two parameters:
+                - geo (str) - containing the two-letter ISO code of the country, e.g. "US", or "" empty string
+                    for global;
+                - timeframe (str) - containing the timeframe in which to query, e.g. "2019-11-28 2020-07-28".
 
-                gtab_config - a dictionary containing values to overrwite some of the configuration parameters for the GTAB methodology. It contains the following parameters:
+            gtab_config - a dictionary containing values to overrwite some of the configuration parameters for the GTAB
+                methodology. It contains the following parameters:
                     - num_anchor_candidates (int) - how many candidates to consider from the anchor candidate list;
                     - num_anchors (int) - how many to sample;
                     - seed (int) - the random seed;
                     - sleep (float) - sleep  e.g. 0.5,
-                    - thresh_offline - the threshold below which the Google Trends query is discarded in the offline phase, e.g. 10.
+                    - thresh_offline - the threshold below which the Google Trends query is discarded in the offline
+                        phase, e.g. 10.
 
-                conn_config - a dictionary containing values to overrwite some of the configuration parameters for the connection. It contains:
+            conn_config - a dictionary containing values to overrwite some of the configuration parameters for the
+                connection. It contains:
                     - backoff_factor (float) - e.g. 0.1;
-                    - proxies (list of strings) - which proxies to use, e.g. ["https://50.2.15.109:8800", "https://50.2.15.1:8800"];
+                    - proxies (list of strings) - which proxies to use,
+                        e.g. ["https://50.2.15.109:8800", "https://50.2.15.1:8800"];
                     - retries (int) - how many times to retry connection;
                     - timeout (list of two values) - e.g. [25, 25]
 
-                overwrite_file - whether to overwrite the config_py.json file in the active directory.
+            overwrite_file - whether to overwrite the config_py.json file in the active directory.
             """
 
         if pytrends_config is not None:
@@ -646,12 +723,11 @@ class GTAB:
 
     def set_blacklist(self, keywords, overwrite_file: bool = False):
         """
-            Sets a new blacklist. This can also be done manually by editing 'config_py.json' in the active directory.
-
-            Input parameters:
-                keywords - list of str keywords;
-                overwrite_file - whether to overwrite the config_py.json file in the active directory.
+        Sets a new blacklist. This can also be done manually by editing 'config_py.json' in the active directory.
 
+        Input parameters:
+            keywords - list of str keywords;
+            overwrite_file - whether to overwrite the config_py.json file in the active directory.
         """
         if type(keywords) != list:
             raise TypeError("Keywords paremeter must be a list with elements of type str.")
@@ -669,12 +745,11 @@ class GTAB:
 
     def set_hitraffic(self, keywords, overwrite_file: bool = False):
         """
-            Sets a new hitraffic list. This can also be done manually by editing 'config_py.json' in the active directory.
-
-            Input parameters:
-                keywords - list of str keywords;
-                overwrite_file - whether to overwrite the config_py.json file in the active directory.
+        Sets a new hitraffic list. This can also be done manually by editing 'config_py.json' in the active directory.
 
+        Input parameters:
+            keywords - list of str keywords;
+            overwrite_file - whether to overwrite the config_py.json file in the active directory.
         """
         if type(keywords) != list:
             raise TypeError("Keywords paremeter must be a list with elements of type str.")
@@ -781,8 +856,11 @@ class GTAB:
         self._log_con = open(os.path.join(self.dir_path, "logs", f"log_{self._make_file_suffix()}.txt"),
                              'a')  # append vs write
         self._log_con.write(f"\n{datetime.datetime.now()}\n")
+
         self._print_and_log(
-            f"Start AnchorBank init for region {self.CONFIG['PYTRENDS']['geo']} in timeframe {self.CONFIG['PYTRENDS']['timeframe']}...")
+        "Start AnchorBank init for region {} in timeframe {}...".format(
+            self.CONFIG['PYTRENDS']['geo'], self.CONFIG['PYTRENDS']['timeframe']
+        ))
 
         if verbose:
             self._print_and_log(f"Full option parameters are: {self.CONFIG}")
@@ -861,7 +939,7 @@ class GTAB:
 
         self._log_con.close()
 
-    def new_query(self, query, first_comparison=None, thresh=100 / np.e, verbose=False, complete=False):
+    def new_query(self, query, first_comparison=None, thresh=10, verbose=False, complete=False):
         """ Request a new Google Trends query and calibrate it with the active gtab.  The 'PYTRENDS' key in
         config_py.json is used. Make sure to set it according to your chosen anchorbank!
 
@@ -889,10 +967,10 @@ class GTAB:
             else list(self.anchor_bank.index).index(first_comparison)
         anchors = tuple(self.anchor_bank.index)
 
-        if query in anchors:
-            self._print_and_log(f"The query is already present in the active gtab!")
-            self._log_con.close()
-            return -2
+        # if query in anchors:
+        #     self._print_and_log(f"The query is already present in the active gtab!")
+        #     self._log_con.close()
+        #     return -2
 
         if not self._check_keyword(query):
             self._print_and_log(f"Keyword {query} is bad!")
@@ -914,12 +992,15 @@ class GTAB:
             except Exception as e:
                 if "response" in dir(e):
                     if e.response.status_code == 429:
-                        input("Quota reached! Please change IP and press any key to continue.")
+                        input("Quota reached! Please change IP and press enter to continue.")
                         ts = self._query_google(keywords=[anchor, query]).iloc[:, 0:2]
                 else:
                     self._print_and_log(f"Google query '{query}' failed because: {str(e)}")
                     break
 
+            if anchor == query:
+                query = "tmp"
+
             timestamps = ts.index
             max_anchor = ts.loc[:, anchor].max()
             max_query = ts.loc[:, query].max()
@@ -938,7 +1019,7 @@ class GTAB:
                 ratio_anchor_lo = self.anchor_bank_lo[anchor]
                 ratio_anchor_hi = self.anchor_bank_hi[anchor]
 
-                #NEW 
+                # NEW
                 ts_query = ts.loc[:, query] / max_anchor * ratio_anchor
                 ts_query_hi = np.array(ts_query_hi) / max_anchor_lo * ratio_anchor_hi
                 ts_query_lo = np.array(ts_query_lo) / max_anchor_hi * ratio_anchor_lo
